#! /usr/bin/perl

    
=head1 NAME

gardol - Adaptive Denial of Service Attack Mitigation

=head1 SYNOPSIS

B<gardol.pl>
[I<options>]
I<HTTP_log_to_monitor>
I<IP_Filter_log_to_monitor>

=head1 DESCRIPTION

B<gardol> is a Perl program which monitors an HTTP log file to
detect accesses which identify IP addresses participating in a
Distributed Denial of Service (DDoS) attack and, in conjunction
with the
S<B<IP Filter>>
package, dynamically block them from attacking a server.


=head1 OPTIONS

All options may be abbreviated to their shortest
unambiguous prefix.

=over 5

=item B<--blacklist> I<filename>

Whenever the list of IP addresses to be blocked is updated, the
IP addresses listed in the specified I<filename> (one per line, with
blank lines and any text following a "#" character ignored) will be
added to the list of IP addresses to be blocked.

=item B<--copyright>

Display copyright information.

=item B<--help>

Display how to call information.


=item B<--ipfconf> I<filename>

The IP Filter configuration file containing the list of IP addresses
to be blocked is written to the specified I<filename>.
The default configuration file name is C<B<ipf.conf>> in
the current directory.  If a I<filename> of "B<->" is specified,
the installation default template location of
C<B</etc/opt/ipf/ipf.conf>> is specified.


=item B<--ipfsentinel> I<string>

The list of IP addresses to be blocked is interpolated into the
B<--ipftemplate> file when a line containing the I<string> is
read.  The default sentinel is:

S<C<#### INSERT ATTACK_DENIAL RULES HERE ####>>


=item B<--ipftemplate> I<filename>

The specified I<filename> is used as the template for the IP
Filter configuration file.  Lines of the template are copied
to the configuration file, with the list of attacking IP
addresses to be blocked interpolated at the specified
B<--ipfsentinel>.
The default template file name is C<B<ipf_conf_template.txt>>
in the current directory.  If a I<filename> of "B<->" is specified,
the installation default template location of
C<B</etc/opt/ipf/ipf_conf_template.txt>> is specified.


=item B<--ipfupdate> I<command>

The specified I<command> will be executed to inform
IP Filter when changes are made in the configuration
file, instructing it to load the new rules.  The default
command is blank, which disables reloading of the filter
rules.  Specifying a I<command> of "B<->" sets the reload command
to the IP Filter installation default of
C<B</etc/init.d/ipfboot reipf>>.


=item B<--minhits> I<n>

A total of I<n> consecutive attack packets (default 2) from a
given host will be required before its IP address is added to
the block list..


=item B<--polltime> I<t>

The HTTP and IP Filter log files will be checked for newly appended
records every I<t> (default 10) seconds.  Log records are read and
processed until the end of the log file is encountered, whereupon
B<gardol> sleeps for I<t> seconds before checking whether additional
items have been added.


=item B<--timeout> I<t>

Hosts from which neither an attack packet nor a packet blocked by
IP Filter has been received in I<t> seconds (default 1800 seconds--30
minutes) are timed out and access from their IP addresses re-enabled.
This prevents floating IP addresses from remaining blocked once an
attacker ceases to hit from them.


=item B<--updtime> I<t>

The IP Filter configuration file is updated, adding host newly added
to the block list and removing those which have timed out, every
I<t> seconds (default 300--five minutes).  The
B<--updtime> should be sent sufficiently short to respond to
the changing population of attacking hosts in a timely
manner, but not so short as to create too much overhead updating
and reloading the filter configuration.


=item B<--verbose>

Generate verbose output to indicate what's going on.


=item B<--version>

Display version number.

=back


=head1 VERSION

This is B<gardol> version 0.1, released March 2004.
The current version of this program is always posted at
http://www.fourmilab.ch/webtools/gardol/.

=head1 AUTHOR

John Walker
(http://www.fourmilab.ch/)

=head1 BUGS

Please report any bugs to bugs@fourmilab.ch.

=head1 SEE ALSO

B<IP Filter> (http://coombs.anu.edu.au/~avalon/),
B<nuweb> (http://nuweb.sourceforge.net/),
S<Literate Programming> (http://www.literateprogramming.com/).

=head1 COPYRIGHT

This program is in the public domain.

=cut


    
    
    require 5;
    use strict;

    
    use Time::Local;
        
    
    
    my $parse_common = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)"/;

    
    my $parse_combined = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)"/;

    
    my $parse_forensic = qr/^(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+(\S+)\s+\[(.*)\]\s+"(.*)"\s+(\d+)\s+([\-\d]+)\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"\s"((?:[^"]|"")*)"/;

    
    
    my $parse_IP_Filter = qr-^(\d+)/(\d+)/(\d+)\s+(\d+):(\d+):(\d+)\.(\d+)\s+\w+\s+\@\d+:\d+\s+(\w)\s+(\d+\.\d+\.\d+\.\d+),-;



    

    #   HTTP transfer log parsing pattern
    my $parseHTTPlog = $parse_forensic;

    #   How long to sleep between scans of new log items
    my $sleepytime = 10;
        
    #   Frequency of update dumps in seconds
    my $updtime = 5 * 60;
    
    #   Time out hosts after this many seconds without a hit
    my $timeout = 30 * 60;
    
    #   Minimum hits from IP address to deem an attacker
    my $minhits = 2;
    
    #   Where to read IP Filter configuration file template
    my $templatefile = 'ipf_conf_template.txt';
    
    #   Sentinel marking where rules are interpolated in ipf.conf file
    my $templatesentinel = '#### INSERT ATTACK_DENIAL RULES HERE ####';
    
    #   Where to read explicit list of IP addresses to blacklist
    my $blacklistfile = '';
    
    #   Where to write IP Filter configuration file
    my $filtfile = 'ipf.conf';
    
    #   Command to load new rule set into IP Filter
    my $reload_IP_Filter = '';


            
    
    #   Processing arguments and options
    
    my $verbose = 0;            # Verbose output for debugging
    
    #   Handy constants
    
    my %mnames = split(/,/, "Jan,1,Feb,2,Mar,3,Apr,4,May,5,Jun,6,Jul,7,Aug,8,Sep,9,Oct,10,Nov,11,Dec,12");
    
    #   Utility variables

    my $lastupd = 0;            # Time of last update dump
    my $restart_signal_received = 0; # Nonzero if HUP restart signal received
    
    my ($l, $ip, $ident, $userid, $time_date, $request, $status, $length, $referer,
           $agent, $cachecont, $pragma, $proxy, $time_date, $mday, $mon, $year,
           $hour, $minute, $second, $timezone, $mindex, $iso_date, $utime, $bp);
           
    my (%hits, %ufirst, %ulast);



    
    
    use Getopt::Long;
        
    GetOptions(
                'blacklist=s' => \$blacklistfile,
                'copyright' => sub { print("This program is in the public domain.\n"); exit(0); },
                'help' => sub { &print_command_line_help; exit(0); },
                'ipfconf=s' => \$filtfile,
                'ipfsentinel=s' => \$templatesentinel,
                'ipftemplate=s' => \$templatefile,
                'ipfupdate=s' => \$reload_IP_Filter,
                'minhits=i' => \$minhits,
                'polltime=i' => \$sleepytime,
                'timeout=i' => \$timeout,
                'updtime=i' => \$updtime,
                'verbose' => \$verbose,
                'version' => sub { print("Version 0.1, March 2004\n"); exit(0); }
              );
              

    

    {
        my $ok = 1;
        
        if (($blacklistfile ne '') && (!(-f $blacklistfile))) {
            print(STDERR "Blacklist file $blacklistfile does not exist.\n");
            $ok = 0;
        }
        
        if ($reload_IP_Filter eq '-') {
            $reload_IP_Filter = '/etc/init.d/ipfboot reipf';
        }
        
        if ($templatesentinel eq '-') {
            $templatesentinel = '/etc/opt/ipf/ipf_conf_template.txt';
        }
        
        if ($templatefile eq '-') {
            $templatefile = '/etc/opt/ipf/ipf.conf';
        }
                
        if (!$ok) {
            die("Invalid option specification(s)");
        }
    }

    
    if ($#ARGV != 1) {
        &print_command_line_help;
        exit(0);
    }
    
    #   HTTP Log file to monitor
    my $read_HTTP_log = $ARGV[0];
    
    #   IP Filter packet disposition log
    my $read_IP_Filter_log = $ARGV[1];
        
    
    open(FL, "<$read_HTTP_log") || die "Unable to open HTTP log file $read_HTTP_log";
    seek(FL,  0, 2);            # Seek to end of log file

    open(FLOG, "<$read_IP_Filter_log") || die "Cannot open IP Filter log $read_IP_Filter_log";
    seek(FLOG,  0, 2);          # Seek to end of IP Filter log file
  ;
    
    
    $SIG{HUP} = sub {
        $restart_signal_received++;
#print("Restart! ($restart_signal_received)\n");
    };

    
    while (1) {
    
        
    if ($restart_signal_received) {
        close(FL);
        close(FLOG);
        
    open(FL, "<$read_HTTP_log") || die "Unable to open HTTP log file $read_HTTP_log";
    seek(FL,  0, 2);            # Seek to end of log file

    open(FLOG, "<$read_IP_Filter_log") || die "Cannot open IP Filter log $read_IP_Filter_log";
    seek(FLOG,  0, 2);          # Seek to end of IP Filter log file
  
        $restart_signal_received = 0;
        if ($verbose) {
            print(STDERR "Log files re-opened.\n");
        }
    }

    
        
    while ($l = <FL>) {
    
        #   Parse request record
        if ($l !~ m/$parseHTTPlog/) {
            #   Can't parse log item.  Ignore it.
            if ($verbose) {
                print(STDERR "!! Cannot parse HTTP log item: $l");
            }
              next; 
        }

        
    #   Special status 573 indicates an attack hit pre-detected by the
    #   filter in Apache.
    if ($6 != 573) {
        #   But if an apparently attacking host sends a single packet
        #   which doesn't resemble an attack, remove it from the list
        #   of attackers.  Its future behaviour will determine whether
        #   we put it back on the list of miscreants.
        if (defined($hits{$1})) {
            
    delete($ulast{$1});
    delete($ufirst{$1});
    delete($hits{$1});

            if ($verbose) {
                print(STDERR "--Purged by non-homepage hit $1\n");
            }
        }
          next; 
    }

    #   Quick reject hits relayed by proxy servers
    if ($12 ne '-') {
          next; 
    }


        
    $ip = $1;
    $ident = $2;
    $userid = $3;
    $time_date = $4;
    $request = $5;
    $status = $6;
    $length = $7;
    $referer = $8;
    $agent = $9;
    $cachecont = $10;
    $pragma = $11;
    $proxy = $12;
    if ($length eq '-') {
        $length = 0;
    }

    #   Parse date and time field
    $time_date =~ m-(\d+)/(\w+)/(\d+):(\d+):(\d+):(\d+)\s([\+\-]\d+)$-;
    $mday = $1;
    $mon = $2;
    $year = $3;
    $hour = $4;
    $minute = $5;
    $second = $6;
    $timezone = $7;
    $mindex = $mnames{$mon};
    $iso_date = sprintf("%04d-%02d-%02d %02d:%02d:%02d", $year, $mindex, $mday,
                            $hour, $minute, $second);
    $utime = timelocal($second, $minute, $hour, $mday, $mindex - 1, $year);
    
#print("$mday,$iso_date,$year,$hour,$minute,$second,$timezone\n");
#print ("$ip,$ident,$userid,$time_date,$request,$status,$length,$cachecont,$pragma,$proxy\n");

        
        
    #   There are no items requiring non-quick rejection here.
    #   An example of a non-quick rejection test is:
    #
    #       if (!(($request =~ m-/earthview/cache/-) && ($status == 404))) {
    #             next; 
    #       }


        
    $hits{$ip}++;
    if ($verbose) {
        print(STDERR "--($hits{$ip}) $l");
    }
    if (!defined($ufirst{$ip})) {
        $ufirst{$ip} = $utime;
    }
    $ulast{$ip} = $utime;

    }

        
        
    while ($l = <FLOG>) {
        if ($l =~ m/$parse_IP_Filter/) {
            $utime = timelocal($6, $5, $4, $1, $2 - 1, $3);
            $bp = $8;
            $ip = $9;
            if (($bp eq 'b') && (defined($ufirst{$ip}))) {
                $ulast{$ip} = $utime;
                $hits{$ip}++;
                if ($verbose) {
                    print(STDERR "-+($hits{$ip}) $l");
                }
            }
        }
    }

        
        if ((time() - $lastupd) >= $updtime) {
            
    my $now = time();

    open(TE, "<$templatefile") || die "Unable to open template file $templatefile";
    open(OF, ">$filtfile") || die "Unable to create IP Filter configuration file $filtfile";

    
    while (<TE>) {
        if ($_ =~ m/^$templatesentinel/) {
            last;
        }
        print(OF);
    }


    
    my @toh;
    my $nblock = 0;
    my $k;

    foreach $k (sort keys %ulast) {
        if (($now - $ulast{$k}) >= $timeout) {
            push(@toh, $k);
        } else {
            if ($hits{$k} >= $minhits) {
                my $mina = int(($now - $ufirst{$k}) / 60);
                print(OF "block in log quick from $k to any  # $mina min. $hits{$k} hits\n");
                $nblock++;
            }
        }
    }

    
    
    if (($blacklistfile ne '') && (-f $blacklistfile)) {
        my $b;
        
        if (open(BL, "<$blacklistfile")) {
            while ($b = <BL>) {
                $b =~ s/#.*$//;
                $b =~ s/^\s*//;
                $b =~ s/\s+$//;
                if ($b ne '') {
                    if ($b =~ s/^\+//) {
                        print(OF "$b  # Transcribed from --blacklist $blacklistfile\n");
                    } else {
                        print(OF "block in log quick from $b to any  # Included from --blacklist $blacklistfile\n");
                    }
                }
            }
            close(BL);
        }
    }


    
    while (<TE>) {
        if ($_ =~ m/^$templatesentinel/) {
            last;
        }
        print(OF);
    }


    
    foreach $k (@toh) {
        
    delete($ulast{$k});
    delete($ufirst{$k});
    delete($hits{$k});

        if ($verbose) {
            print(STDERR "--Timeout purged $k\n");
        }
    }


    close(OF);
    if ($verbose) {
        printf(STDERR "--Blocked $nblock hosts.\n");
    }
    $lastupd = time();

    if ($reload_IP_Filter ne '') {
        system("$reload_IP_Filter\n");
    }

        }
        if ($verbose) {
            print(STDERR "--Sleeping $sleepytime seconds.\n");
        }
        sleep($sleepytime);
    }
             
    
    
    sub print_command_line_help {
        print << "EOD";
Usage: gardol.pl [ options ] HTTP_log_to_monitor IP_Filter_log_to_monitor
       Options:
             --blacklist f   Add IP addresses in blacklist file f to block list
             --copyright     Print copyright information
             --help          Print this message
             --ipfconf f     Write IP Filter configuration to file f
             --ipfsentinel s Interpolate host blocking commands at this sentinel in IP Filter template
             --ipftemplate f Read IP Filter configuration template from file f
             --ipfupdate c   Use shell command c to update IP Filter configuration
             --minhits n     Require n consecutive attack packets before blocking host
             --polltime n    Poll log files for new items every n seconds
             --timeout n     Timeout and unblock hosts after n seconds of inactivity
             --updtime n     Update list of blocked sites every n seconds
             --verbose       Print verbose debugging information
             --version       Print version number
Version 0.1, March 2004
EOD
   }


